import json
from typing import Dict, List, Optional, Union

from rich import box
from rich.console import Console
from rich.table import Table

console = Console()


def __table(
    data: Union[Dict, List],
    columns: List[str] = None,
    title: Optional[str] = None,
):
    """
    Print a table from a JSON response using rich.

    Parameters:
        data (dict or list): The JSON data.
        path (list): Path to the list of items you want to tabulate.
        columns (list): The keys from each item to show as columns.
        title(str)
    """

    if isinstance(data, list):
        if not columns:
            columns = list(data[0].keys())
        rows = data
    elif isinstance(data, dict):
        if not columns:
            columns = list(data.keys())
        rows = [data]
    else:
        return

    table = Table(title=title, box=box.SIMPLE)

    for col in columns:
        table.add_column(col, style="cyan", no_wrap=True)

    for item in rows:
        row = [str(item.get(col, "")) for col in columns]
        table.add_row(*row)

    console.print(table)


def __json(data):
    console.print_json(json.dumps(data))


def raw(data):
    console.print(data)


def print(
    args,
    data: Union[Dict, List],
    columns: List[str] = None,
    title: Optional[str] = None,
):
    if args.json:
        __json(data)
    else:
        __table(data=data, columns=columns, title=title)



#!/usr/bin/env python3


class InvalidCommand(Exception):
    def __init__(self, msg):
        self.msg = msg


class RequestError(Exception):
    def __init__(self, url, status, body):
        self.url = url
        self.status = status
        self.body = body



#!/usr/bin/env python

import argparse
import json
import logging
import sys
from os import getenv

from rich import print
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich_argparse import ArgumentDefaultsRichHelpFormatter

from . import errors, models, printer
from .api import Portainer
from .client import Client


def configure_logging(args):
    if args.debug:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
            datefmt="[%X]",
            handlers=[RichHandler(rich_tracebacks=True)],
        )


def subcommand(parser: argparse._SubParsersAction, name: str, **kwargs):
    return parser.add_parser(
        name, formatter_class=ArgumentDefaultsRichHelpFormatter, **kwargs
    )


def parse_mount(conf: str):
    try:
        path, name = conf.split(":")
        return (path, name)
    except Exception:
        raise errors.InvalidCommand("invalid mount argument: " + conf)


def name_or_id_group(parser: argparse.ArgumentParser):
    args = parser.add_mutually_exclusive_group(required=True)
    args.add_argument("-n", "--name")
    args.add_argument("--id")


def docker_filter_group(parser: argparse.ArgumentParser):
    args = parser.add_argument_group()
    args.add_argument(
        "-n",
        "--name",
        action="append",
        default=[],
    )
    args.add_argument(
        "-l",
        "--label",
        action="append",
        default=[],
    )
    args.add_argument(
        "--id",
        action="append",
        default=[],
    )


def file_or_inline(parser: argparse.ArgumentParser):
    args = parser.add_mutually_exclusive_group(required=True)
    args.add_argument("-f", "--file")
    args.add_argument("--value")


def requires_endpoint(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-E",
        "--endpoint",
        help="id of the endpoint to deploy on",
        required=True,
    )


def read_string(path):
    with open(path, "r") as f:
        return "".join(f.readlines())


def get_unauthenticated_api(args):
    client = Client(args.host)
    return Portainer(client)


def get_authenticated_api(args):
    HOST = getenv("PORTAINER_HOST", args.host)
    PASSWORD = getenv("PORTAINER_PASSWORD", args.password)
    USERNAME = getenv("PORTAINER_USERNAME", args.username)
    API_TOKEN = getenv("PORTAINER_TOKEN", args.api_token)

    client = Client(HOST)
    if API_TOKEN:
        client.authorize(API_TOKEN)
    else:
        client.login(USERNAME, PASSWORD)
    return Portainer(client)


def deploy(args):
    request = models.DeploymentRequest()
    api = get_authenticated_api(args)

    request.name = args.stack_name

    with open(args.compose_file, "r") as f:
        request.compose = "".join(f.readlines())

    env_lines = args.variable
    if args.env_file != None:
        with open(args.env_file, "r") as f:
            lines = f.readlines()
            env_lines = env_lines + lines

    request.variables = dict()
    for line in env_lines:
        striped = line.strip()
        idx = striped.find("=")
        request.variables[striped[:idx]] = striped[idx + 1 :]

    request.configs = dict(
        [
            (name, read_string(path))
            for (path, name) in [parse_mount(conf) for conf in args.config]
        ]
    )

    request.secrets = dict(
        [
            (name, read_string(path))
            for (path, name) in [parse_mount(conf) for conf in args.secret]
        ]
    )

    return api.endpoint(args.endpoint).deploy(request, args.no_error)


def _build_deploy_cmd(subparsers):
    deploy_cmd = subcommand(subparsers, "deploy")
    deploy_cmd.add_argument(
        "-f",
        "--compose-file",
        help="compose manifest file",
        default="docker-compose.yaml",
        required=True,
    )
    requires_endpoint(deploy_cmd)
    deploy_cmd.add_argument(
        "-S",
        "--stack-name",
        help="Name of the stack to create or update",
        required=True,
    )
    deploy_cmd.add_argument(
        "--env-file",
        help="dot env file used for deployment, it will be used as stack environment in portainer",
    )
    deploy_cmd.add_argument(
        "-e",
        "--variable",
        action="append",
        help="environment variable `SOME_ENV=some-value`",
        default=[],
    )
    deploy_cmd.add_argument(
        "-c",
        "--config",
        help="""create config; args must be like `local-path-to-file:conf-name`;
  NOTE that as configs are immutable and might be already in use, your config name must not exist!
  use versioning or date in names to always get a new name
  """,
        action="append",
        default=[],
    )
    deploy_cmd.add_argument(
        "-s",
        "--secret",
        help="create a new secret; see --config.",
        action="append",
        default=[],
    )

    deploy_cmd.add_argument(
        "-K",
        "--no-error",
        help="Don't raise error if there are existing configs or secrets",
        action="store_true",
    )
    deploy_cmd.set_defaults(func=deploy)


def _build_endpoints_cmd(subparsers):
    endpoints_cmd = subcommand(subparsers, "endpoints")
    endpoints_cmd.set_defaults(func=lambda args: endpoints_cmd.print_help())

    subcmd = endpoints_cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    def get(args):
        client = get_authenticated_api(args)
        if args.name:
            printer.print(args, data=client.endpoints.get_by_name(args.name))
        elif args.id:
            printer.print(args, data=client.endpoints.get(args.id))

    def ls(args):
        client = get_authenticated_api(args)

        printer.print(
            args,
            data=client.endpoints.list(),
            columns=["Id", "Name", "Type", "URL", "ContainerEngine"],
        )

    def info(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.endpoint(args.endpoint).get_docker_info())

    def version(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.endpoint(args.endpoint).get_docker_version())

    def create(args):
        client = get_authenticated_api(args)
        request = models.EndpointCreationRequest()
        request.name = args.name
        request.tagIds = args.tag
        request.groupId = args.group
        request.url = args.url
        request.type = args.type

        printer.print(args, data=client.endpoints.create(request))

    get_cmd = subcommand(subcmd, "get")
    name_or_id_group(get_cmd)
    get_cmd.set_defaults(func=get)

    ls_cmd = subcommand(subcmd, "ls")
    ls_cmd.set_defaults(func=ls)

    create_cmd = subcommand(subcmd, "create")
    create_cmd.add_argument("name", help="Endpoint name")
    create_cmd.add_argument(
        "--type",
        choices=list(models.EndpointCreationType),
        type=models.EndpointCreationType.from_string,
    )
    create_cmd.add_argument("-u", "--url")
    create_cmd.add_argument(
        "-t",
        "--tag",
        action="append",
        default=[],
    )
    create_cmd.add_argument("-g", "--group")
    create_cmd.set_defaults(func=create)

    info_cmd = subcommand(subcmd, "info")
    requires_endpoint(info_cmd)
    info_cmd.set_defaults(func=info)

    version_cmd = subcommand(subcmd, "version")
    requires_endpoint(version_cmd)
    version_cmd.set_defaults(func=version)


def _build_stacks_cmd(subparsers):
    stacks_cmd = subcommand(subparsers, "stacks")
    stacks_cmd.set_defaults(func=lambda args: stacks_cmd.print_help())

    subcmd = stacks_cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    def get(args):
        client = get_authenticated_api(args)
        if args.name:
            printer.print(args, data=client.stacks.get_stacks_by_name(args.name))
        elif args.id:
            printer.print(args, data=client.stacks.get(args.id))

    def get_file(args):
        client = get_authenticated_api(args)
        printer.raw(client.stacks.get_file(args.id))

    def ls(args):
        client = get_authenticated_api(args)
        printer.print(
            args,
            data=client.stacks.list(),
            columns=["Id", "Name", "Type", "Status", "EndpointId", "SwarmId"],
        )

    def delete(args):
        client = get_authenticated_api(args)
        if args.id:
            printer.print(
                args,
                client.endpoint(args.endpoint).stacks.delete(args.id, args.external),
            )
        elif args.name:
            printer.print(
                args,
                client.endpoint(args.endpoint).stacks.delete_by_name(
                    args.name, args.external
                ),
            )

    def start(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.endpoint(args.endpoint).stacks.start(args.id))

    def stop(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.endpoint(args.endpoint).stacks.stop(args.id))

    _build_deploy_cmd(subcmd)

    ls_cmd = subcommand(subcmd, "ls")
    ls_cmd.set_defaults(func=ls)

    get_cmd = subcommand(subcmd, "get")
    name_or_id_group(get_cmd)
    get_cmd.set_defaults(func=get)

    get_file_cmd = subcmd.add_parser("get-file")
    get_file_cmd.add_argument("id")
    get_file_cmd.set_defaults(func=get_file)

    start_cmd = subcommand(subcmd, "start")
    start_cmd.add_argument("id")
    requires_endpoint(start_cmd)
    start_cmd.set_defaults(func=start)

    stop_cmd = subcommand(subcmd, "stop")
    stop_cmd.add_argument("id")
    requires_endpoint(stop_cmd)
    stop_cmd.set_defaults(func=stop)

    delete_cmd = subcommand(subcmd, "delete")
    name_or_id_group(delete_cmd)
    requires_endpoint(delete_cmd)
    delete_cmd.add_argument(
        "--external",
        help="Whether or not delete if it is an external stack",
        action="store_true",
    )
    delete_cmd.set_defaults(func=delete)


def _build_tags_cmd(subparsers):
    tags_cmd = subcommand(subparsers, "tags")
    tags_cmd.set_defaults(func=lambda args: tags_cmd.print_help())

    subcmd = tags_cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    def create(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.tags.create(args.name))

    def ls(args):
        client = get_authenticated_api(args)
        printer.print(args, data=client.tags.list())

    def delete(args):
        client = get_authenticated_api(args)
        client.tags.delete(args.id)

    create_cmd = subcommand(subcmd, "create")
    create_cmd.add_argument("name")
    create_cmd.set_defaults(func=create)

    ls_cmd = subcommand(subcmd, "ls")
    ls_cmd.set_defaults(func=ls)

    delete_cmd = subcommand(subcmd, "delete")
    delete_cmd.add_argument("id")
    delete_cmd.set_defaults(func=delete)


def _build_secrets_cmd(subparsers):
    secrets_cmd = subcommand(subparsers, "secrets")
    secrets_cmd.set_defaults(func=lambda args: secrets_cmd.print_help())

    subcmd = secrets_cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    requires_endpoint(secrets_cmd)

    def create(args):
        client = get_authenticated_api(args)
        data = args.value if args.value else read_string(args.file)
        printer.print(
            args, client.endpoint(args.endpoint).secrets.create(args.name, data)
        )

    def ls(args):
        client = get_authenticated_api(args)
        printer.print(
            args,
            client.endpoint(args.endpoint).secrets.ls(
                id=args.id, name=args.name, label=args.label
            ),
        )

    def delete(args):
        client = get_authenticated_api(args)
        client.endpoint(args.endpoint).secrets.delete(args.id)

    create_cmd = subcommand(subcmd, "create")
    create_cmd.add_argument("name")
    file_or_inline(create_cmd)
    create_cmd.set_defaults(func=create)

    ls_cmd = subcommand(subcmd, "ls")
    docker_filter_group(ls_cmd)
    ls_cmd.set_defaults(func=ls)

    delete_cmd = subcommand(subcmd, "delete")
    delete_cmd.add_argument("id")
    delete_cmd.set_defaults(func=delete)


def _build_configs_cmd(subparsers):
    configs_cmd = subcommand(subparsers, "configs")
    configs_cmd.set_defaults(func=lambda args: configs_cmd.print_help())

    subcmd = configs_cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    requires_endpoint(configs_cmd)

    def create(args):
        client = get_authenticated_api(args)
        data = args.value if args.value else read_string(args.file)
        printer.print(
            args, client.endpoint(args.endpoint).configs.create(args.name, data)
        )

    def ls(args):
        client = get_authenticated_api(args)
        printer.print(
            args,
            client.endpoint(args.endpoint).configs.ls(
                id=args.id, name=args.name, label=args.label
            ),
        )

    def delete(args):
        client = get_authenticated_api(args)
        client.endpoint(args.endpoint).configs.delete(args.id)

    create_cmd = subcommand(subcmd, "create")
    create_cmd.add_argument("name")
    file_or_inline(create_cmd)
    create_cmd.set_defaults(func=create)

    ls_cmd = subcommand(subcmd, "ls")
    docker_filter_group(ls_cmd)
    ls_cmd.set_defaults(func=ls)

    delete_cmd = subcommand(subcmd, "delete")
    delete_cmd.add_argument("id")
    delete_cmd.set_defaults(func=delete)


def _build_system_cmd(subparsers):
    cmd = subcommand(subparsers, "system")
    cmd.set_defaults(func=lambda args: cmd.print_help())

    subcmd = cmd.add_subparsers(
        title="supported resources", help="resource to get info"
    )

    def build_init_cmd():
        def init(args):
            PASSWORD = getenv("PORTAINER_PASSWORD", args.password)
            USERNAME = getenv("PORTAINER_USERNAME", args.username)
            if not (USERNAME and PASSWORD):
                raise errors.InvalidCommand(
                    "Both username and password are required for admin initiation."
                )
            portainer = get_unauthenticated_api(args)

            printer.print(
                args, portainer.public.init(username=USERNAME, password=PASSWORD)
            )

        init_cmd = subcommand(subcmd, "init")
        init_cmd.set_defaults(func=init)

    def build_status_cmd():
        def get(args):
            client = get_unauthenticated_api(args)
            printer.print(args, data=client.public.status())

        status = subcommand(subcmd, "status")
        status.set_defaults(func=get)

    build_init_cmd()
    build_status_cmd()


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Poorman's kubectl, CLI for portainer on docker swarm",
        epilog=Markdown(
            """No budget. No vendors. No fleet of ops.
        Just you, a blinking cursor, and the will to script what others buy.
        The rich scale with dollars. You scale with shell.
        Excuses cost, Automation pays! 🔧
        """,
            style="italic sky_blue1",
        ),
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    ### HACK: this is due to a known issue in argparse in python3
    parser.set_defaults(func=lambda args: parser.print_help())

    parser.add_argument(
        "-T",
        "--api-token",
        help="api token for user, overrides PORTAINER_TOKEN variable",
    )
    parser.add_argument(
        "-H",
        "--host",
        help="portainer host, overrides PORTAINER_HOST variable",
        default="http://127.0.0.1:9000/api",
    )
    parser.add_argument(
        "-U",
        "--username",
        help="username to login, overrides PORTAINER_USERNAME variable",
        default="admin",
    )
    parser.add_argument(
        "-P",
        "--password",
        help="password for user, overrides PORTAINER_PASSWORD variable",
        default="admin",
    )

    parser.add_argument(
        "--debug", help="Whether or not print debugging logs", action="store_true"
    )

    parser.add_argument("-j", "--json", help="Print json output", action="store_true")

    subparsers = parser.add_subparsers(title="commands")

    _build_deploy_cmd(subparsers)
    _build_stacks_cmd(subparsers)
    _build_configs_cmd(subparsers)
    _build_secrets_cmd(subparsers)
    _build_endpoints_cmd(subparsers)
    _build_tags_cmd(subparsers)
    _build_system_cmd(subparsers)

    return parser


def main():
    parser = _build_parser()
    args = parser.parse_args()
    configure_logging(args)
    try:
        args.func(args)
    except errors.InvalidCommand as err:
        print(err.msg, file=sys.stderr)
        sys.exit(1)
    except errors.RequestError as err:
        print(err.url, err.status, err.body, file=sys.stderr)
        sys.exit(2)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import json
import logging

import requests
from requests.structures import CaseInsensitiveDict

from . import errors, helpers

logger = logging.getLogger(__name__)


class Client:
    def __init__(self, host: str):
        self.host = host
        self.token: CaseInsensitiveDict = CaseInsensitiveDict()

    def __extract(self, resp):
        if resp.ok:
            return resp.json()
        else:
            raise errors.RequestError(resp.url, resp.status_code, resp.text)

    def get(self, url: str, **kwargs):
        resp = requests.get(self.host + url, headers=self.token, **kwargs)
        return self.__extract(resp)

    def post(self, url: str, json=None, data=None, **kwargs):
        resp = requests.post(
            self.host + url, headers=self.token, json=json, data=data, **kwargs
        )
        return self.__extract(resp)

    def put(self, url: str, data, **kwargs):
        resp = requests.put(
            self.host + url, headers=self.token, data=json.dumps(data), **kwargs
        )
        return self.__extract(resp)

    def delete(self, url: str, **kwargs):
        resp = requests.delete(self.host + url, headers=self.token, **kwargs)
        if resp.ok:
            return
        else:
            raise errors.RequestError(resp.url, resp.status_code, resp.text)

    def authorize(self, api_token: str):
        logger.info("Authorized for " + self.host + " using api token")
        self.token["Accept"] = "application/json"
        self.token["x-api-key"] = api_token
        return

    def login(self, username: str, password: str):
        logger.info("Trying to login to " + self.host + "...")
        body = {"username": username, "password": password}
        resp = self.post("/auth", body)
        token = resp["jwt"]
        self.token["Accept"] = "application/json"
        self.token["Authorization"] = "Bearer " + token
        return

    def logout(self):
        self.token.clear()



from enum import Enum
from typing import Dict, List, Optional


class EndpointCreationType(Enum):
    LocalDocker = 1
    Agent = 2

    def __str__(self):
        return self.name

    @staticmethod
    def from_string(s):
        try:
            return EndpointCreationType[s]
        except KeyError:
            raise ValueError()


class EndpointCreationRequest:
    name: str
    type: EndpointCreationType
    url: Optional[str]
    tagIds: List[int]
    groupId: int


class DeploymentRequest:
    name: str
    compose: str
    configs: Dict[str, str]
    secrets: Dict[str, str]
    variables: Dict[str, str]



import json
import logging

import requests
from requests.structures import CaseInsensitiveDict

from . import errors, helpers, models
from .client import Client

logger = logging.getLogger(__name__)


class GlobalStacksAPI:
    def __init__(self, client: Client):
        self.__client = client

    def list(self, **kwargs):
        logger.info("Getting a list of stacks")
        return self.__client.get("/stacks", params=kwargs)

    def get_stack_id_by_name(self, name):
        stacks = self.list(name=name)
        ids = [s["Id"] for s in stacks if s["Name"] == name]
        return ids[0]

    def get_stacks_by_name(self, name):
        stacks = self.list()
        filtered = [s for s in stacks if s["Name"] == name]
        return filtered

    def get(self, id):
        logger.info(f"Getting stack {id}")
        return self.__client.get(f"/stacks/{id}")

    def get_file(self, id):
        resp = self.__client.get(f"/stacks/{id}/file")
        return resp["StackFileContent"]


class StacksAPI:
    def __init__(self, client, endpoint_id):
        self.__client = client
        self.__endpoint_id = str(endpoint_id)

    def create(self, stack_name: str, compose: str, env_vars):
        stack_name = stack_name.lower().strip()

        stacks = GlobalStacksAPI(self.__client).get_stacks_by_name(stack_name)

        if len(stacks) == 1:
            logger.info("Updating existing stack name: " + stack_name)
            existing_stack = stacks[0]
            stack_id = existing_stack["Id"]

            data = {
                "Prune": True,
                "StackFileContent": compose,
                "Env": env_vars,
                "id": stack_id,
            }

            return self.__client.put(
                f"/stacks/{str(stack_id)}",
                params={"endpointId": self.__endpoint_id},
                data=data,
            )

        else:
            logger.info("No existing stack with name: " + stack_name)
            info = EndpointAPI(self.__client, self.__endpoint_id).get_docker_info()
            swarm_id = info["Swarm"]["Cluster"]["ID"]

            data = {
                "Env": env_vars,
                "Name": stack_name,
                "SwarmID": swarm_id,
                "StackFileContent": compose,
            }
            return self.__client.post(
                f"/stacks/create/swarm/string",
                params={"endpointId": self.__endpoint_id},
                json=data,
            )

    def start(self, id):
        return self.__client.post(
            f"/stacks/{id}/start", params={"endpointId": self.__endpoint_id}
        )

    def stop(self, id):
        return self.__client.post(
            f"/stacks/{id}/stop",
            params={"endpointId": self.__endpoint_id},
        )

    def delete(self, id, external: bool = False):
        return self.__client.delete(
            f"/stacks/{id}",
            params={
                "endpointId": self.__endpoint_id,
                "external": str(external).lower(),
            },
        )

    def delete_by_name(self, name: str, external: bool = False):
        return self.__client.delete(
            f"/stacks/name/{name}",
            params={
                "endpointId": self.__endpoint_id,
                "external": str(external).lower(),
            },
        )


class ConfigsAPI:
    def __init__(self, client, endpoint_id, api_version: str):
        self.__client = client
        self.__endpoint_id = str(endpoint_id)
        self.__base = (
            f"/endpoints/{self.__endpoint_id}/docker/"
            + (f"v{api_version}/" if api_version else "")
            + "configs"
        )

    def create(self, name, data):
        name = name.strip()

        logger.info("Creating new config " + name)
        body = {"Data": helpers.to_base64(data), "Name": name, "Labels": {}}
        return self.__client.post(f"{self.__base}/create", body)

    def get(self, id: str):
        return self.ls(id=[id])

    def get_by_name(self, name: str):
        return self.ls(name=[name])

    def ls(self, **kwargs):
        return self.__client.get(
            self.__base,
            params={"filters": json.dumps(kwargs)},
        )

    def delete(self, id: str):
        return self.__client.delete(
            f"{self.__base}/{id}",
        )


class SecretsAPI:
    def __init__(self, client, endpoint_id, api_version: str):
        self.__client = client
        self.__endpoint_id = str(endpoint_id)
        self.__base = (
            f"/endpoints/{self.__endpoint_id}/docker/"
            + (f"v{api_version}/" if api_version else "")
            + "secrets"
        )

    def create(self, name, data):
        name = name.strip()

        logger.info("Creating new secret " + name)
        body = {"Data": helpers.to_base64(data), "Name": name, "Labels": {}}
        return self.__client.post(f"{self.__base}/create", body)

    def ls(self, **kwargs):
        return self.__client.get(
            self.__base,
            params={"filters": json.dumps(kwargs)},
        )

    def delete(self, id: str):
        return self.__client.delete(
            f"{self.__base}/{id}",
        )


class EndpointAPI:
    def __init__(self, client, endpoint_id):
        self.__client = client
        self.__endpoint_id = str(endpoint_id)
        logger.info("Getting endpoint info")
        version = self.__client.get(
            "/endpoints/" + self.__endpoint_id + "/docker/version"
        )
        api_version = version["ApiVersion"]
        self.__api_version = api_version
        self.configs = ConfigsAPI(client, endpoint_id, api_version)
        self.secrets = SecretsAPI(client, endpoint_id, api_version)
        self.stacks = StacksAPI(client, endpoint_id)

    def get_docker_info(self):
        logger.info("Getting endpoint info")
        resp = self.__client.get("/endpoints/" + self.__endpoint_id + "/docker/info")
        return resp

    def get_docker_version(self):
        logger.info("Getting endpoint docker version")
        resp = self.__client.get("/endpoints/" + self.__endpoint_id + "/docker/version")
        return resp

    def deploy(self, request: models.DeploymentRequest, no_error: bool = False):
        """
        Deploys a compose file with it's required configs and secrets

        Raises an error if there is at least one already existing config or secret and `no_error` is False
        """

        for name, data in request.configs.items():
            try:
                self.configs.create(name=name, data=data)
            except errors.RequestError:
                if not no_error:
                    raise

        for name, data in request.secrets.items():
            try:
                self.secrets.create(name=name, data=data)
            except errors.RequestError:
                if not no_error:
                    raise

        return self.stacks.create(
            stack_name=request.name,
            compose=request.compose,
            env_vars=[
                {"name": key, "value": value}
                for (key, value) in request.variables.items()
            ],
        )


class EndpointsAPI:
    def __init__(self, client):
        self.__client = client

    def create(self, request: models.EndpointCreationRequest):
        data = {
            "Name": request.name,
            "EndpointCreationType": request.type.value,
            "URL": request.url,
            "GroupID": request.groupId,
            "TagIds": request.tagIds,
        }
        return self.__client.post("/endpoints", data=data)

    def list(self, **kwargs):
        logger.info("Getting a list of endpoints")
        resp = self.__client.get("/endpoints", **kwargs)
        return resp

    def get(self, id):
        logger.info(f"Getting endpoint {id}")
        resp = self.__client.get(f"/endpoints/{id}")
        return resp

    def get_by_name(self, name):
        return self.list(params={"name": name})


class TagsAPI:
    def __init__(self, client):
        self.__client = client

    def create(self, name: str):
        return self.__client.post("/tags", {"name": name})

    def list(self):
        logger.info("Getting a list of tags")
        return self.__client.get("/tags")

    def delete(self, id):
        return self.__client.delete(f"/tags/{id}")

    def get_endpoints(self, tag):
        tags = self.list()
        target = [ei for t in tags for ei in t["Endpoints"] if t["Name"] == tag]
        return target


class PublicAPI:
    def __init__(self, client: Client):
        self.__client = client

    def status(self):
        return self.__client.get("/status")

    def init(self, username: str, password: str):
        data = {"password": password, "username": username}
        return self.__client.post("/users/admin/init", data)


class Portainer:
    def __init__(self, client: Client):
        self.__client = client
        self.stacks = GlobalStacksAPI(client)
        self.endpoints = EndpointsAPI(client)
        self.tags = TagsAPI(client)
        self.public = PublicAPI(client)

    def endpoint(self, id) -> EndpointAPI:
        return EndpointAPI(self.__client, id)
